from t8s.log_config import LogConfig
from pathlib import Path
from datetime import datetime
import numpy as np
import pandas as pd
import t8s
from t8s.ts import TimeSerie
from t8s.ts_builder import TSBuilder, ReadParquetFile, ReadCsvFile
from t8s.log_config import LogConfig
from t8s.ts_writer import TSWriter, WriteParquetFile, WriteCsvFile
from logging import INFO, DEBUG, WARNING, ERROR, CRITICAL

if __name__ == "__main__":
    # print('globals:', globals())
    # The client code.
    LogConfig().initialize_logger(INFO)
    logger = LogConfig().getLogger()
    ts = TSBuilder.empty()
    # initialize_logger(INFO)
    logger.info('t8s package version:' + t8s.__version__)
    path_str: str = 'ts_01.parquet'
    path = Path(path_str)
    # Cria uma série temporal multivariada com três atributos: timestamp, temperatura e velocidade
    data = {
        'timestamp': [
            datetime(2022, 1, 1, 0, 0, 0), 
            datetime(2022, 1, 1, 1, 0, 0),
            datetime(2022, 1, 1, 2, 0, 0), 
            datetime(2022, 1, 1, 3, 0, 0)],
        'temperatura': np.array([25.0, 26.0, 27.0, 23.2], dtype=np.float32),
        'velocidade': [2000, 1100, 1200, 4000]
    }
    # Convertendo os tipos de dado para temperatura e velocidade para 
    # np.float32 e np.int32 respectivamente, pois o padrão é np.float64 e np.int64
    data['temperatura'] = np.array(data['temperatura'], dtype=np.float32)
    data['velocidade'] = np.array(data['velocidade'], dtype=np.int32)
    # Cria uma série temporal multivariada com três atributos: timestamp, temperatura e
    # velocidade para o proposito de teste
    ts = TimeSerie(data, format='wide', features_qty = 3)
    cols_str = [name for name in sorted(ts.df.columns)]
    cols_str = ', '.join(cols_str)
    logger.info(f'Dataframe com {len(ts.df.columns)} colunas: {cols_str}')
    # Faz o display da série temporal no terminal
    print(ts)
    # Grava a série temporal em parquet
    print(f'Grava a série temporal (formato {ts.format}) em um arquivo parquet {path}')
    context = TSWriter(WriteParquetFile())
    print("Client: Strategy was seted to write Parquet file.")
    context.write(Path(path_str), ts)
    print(f'\nArquivo {str(path)} gerado à partir da TimeSerie fornecida')
    # --------------------------------------------------------------------------------
    # Limpando objeto ts para garantir que será lido corretamente.
    ts: TimeSerie = TSBuilder.empty()
    # Lê a série temporal gravada no arquivo parquet
    print(f'\nLendo path {str(path)} e gerando TimeSerie')
    
    # The client code picks a concrete strategy and passes it to the context.
    # The client should be aware of the differences between strategies in order
    # to make the right choice.
    assert isinstance(path, Path), "path must be a Path object"
    if  (str(path)).endswith('.parquet'):
        context = TSBuilder(ReadParquetFile())
        print("Client: ReadStrategy is set to read Parquet file.")
        ts = context.build_from_file(Path(path_str))
    else:
        assert str(path).endswith('.csv'), "If path is not a Parquet file the path must be a CSV file"
        print("Client: ReadStrategy is set to read CSV file.")
        context = TSBuilder(ReadCsvFile())
        ts = context.build_from_file(Path(path_str))

    assert int(ts.features) == 3
    assert ts.format == 'wide'
    assert len(ts.df) == 4
    print('Tipos das colunas no Dataframe da Série Temporal:')
    for col in ts.df.columns:
        print(col, '\t', type(ts.df[col][0]))

    # pd._libs.tslibs.timestamps.Timestamp é privado e devo usar pd.Timestamp
    assert type(ts.df['timestamp'][0]) == pd.Timestamp
    assert type(ts.df['temperatura'][0]) == np.float32
    assert type(ts.df['velocidade'][0]) == np.int32

    univariate_list = ts.split()
    for idx, ts_uni in enumerate(univariate_list):
        print(f'TimeSerie univariada {idx}:')
        print(ts_uni)
        print('---------------------------------------------------')
        print(f'univariate {idx-1}', '\n ', ts_uni)

    print('---------------------------------------------------')
    


# SPDX-FileCopyrightText: 2023-present João Antonio Ferreira <joao.parana@gmail.com>
#
# SPDX-License-Identifier: MIT

import yaml
from pathlib import Path
from datetime import datetime
from t8s.ts import TimeSerie




# SPDX-FileCopyrightText: 2023-present João Antonio Ferreira <joao.parana@gmail.com>
#
# SPDX-License-Identifier: MIT
from .__about__ import __version__
# from abc import ABC, abstractmethod
#
# class ITimeSerie(ABC):
#    pass


# from __future__ import annotations usado apenas nas versões anteriores a 3.7
from abc import ABC, abstractmethod
from typing import Any, Optional
from pathlib import Path
from datetime import datetime
from t8s.ts import TimeSerie
from t8s.log_config import LogConfig
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

logger = LogConfig().getLogger()

class Strategy(ABC):
    """
    The Strategy interface declares operations common to all supported versions
    of some algorithm.

    The Context uses this interface to call the algorithm defined by Concrete
    Strategies.
    """

    @abstractmethod
    def do_write(self, data: Any, ts: Any) -> Any:
        pass


"""
Concrete Strategies implement the algorithm while following the base Strategy
interface. The interface makes them interchangeable in the Context.
"""

class TSWriter():
    """
    The Context defines the interface of interest to clients.
    """

    def __init__(self, strategy: Strategy) -> None:
        """
        Usually, the Context accepts a strategy through the constructor, but
        also provides a setter to change it at runtime.
        """
        logger.info('TSBuilder constructor received strategy: %s', strategy)
        self._strategy = strategy

    @property
    def strategy(self) -> Strategy:
        """
        The Context maintains a reference to one of the Strategy objects. The
        Context does not know the concrete class of a strategy. It should work
        with all strategies via the Strategy interface.
        """

        return self._strategy

    @strategy.setter
    def strategy(self, strategy: Strategy) -> None:
        """
        Usually, the Context allows replacing a Strategy object at runtime.
        """

        self._strategy = strategy

    def write(self, path: Path, ts: TimeSerie) -> None:
        """
        The Context delegates some work to the Strategy object instead of
        implementing multiple versions of the algorithm on its own.
        """

        # ...

        logger.info("Context: write")
        result = self._strategy.do_write(path, ts)
        # logger.info('result type from do_write() is: ' + str(type(result)))

        # ...
        return result

class WriteParquetFile(Strategy):
    # O método do_write é VOID, pois o resultado é gravado em disco. Caso ocorra
    # algum problema uma `Exception` é lançada.
    def do_write(self, path: Path, ts: TimeSerie) -> None:
        logger.info('Using WriteParquetFile strategy')
        # Grava os dados em formato Parquet com metadados do objeto TimeSerie
        # to_parquet(path, df, self.format, self.features)
        table = pa.Table.from_pandas(ts.df)
        # table = table.replace_schema_metadata({'format': self.format, 'features': self.features})
        table = table.replace_schema_metadata({
            b'format': str(ts.format).encode(),
            b'features': str(ts.features).encode()})
        result = pq.write_table(table, path)
        return result      

class WriteCsvFile(Strategy):
    # O método do_write é VOID, pois o resultado é gravado em disco. Caso ocorra
    # algum problema uma `Exception` é lançada.
    def do_write(self, path: Path, ts: TimeSerie) -> None:
        logger.info('Using WriteCsvFile strategy')
        return None



import yaml
from pathlib import Path
from datetime import datetime
import numpy as np
import pandas as pd
from t8s.ts import TimeSerie
from t8s.ts_writer import TSWriter, WriteParquetFile

def test_to_parquet():
    path_str: str = 'ts_01.parquet'
    path = Path(path_str)
    # Cria uma série temporal multivariada com três atributos: timestamp, temperatura e velocidade
    data = {
        'timestamp': [
            datetime(2022, 1, 1, 0, 0, 0), 
            datetime(2022, 1, 1, 1, 0, 0),
            datetime(2022, 1, 1, 2, 0, 0), 
            datetime(2022, 1, 1, 3, 0, 0)],
        'temperatura': [25.0, 26.0, 27.0, 23.2],
        'velocidade': [2000, 1100, 1200, 4000]
    }
    # Convertendo os tipos de dado para temperatura e velocidade para 
    # np.float32 e np.int32 respectivamente, pois o padrão é np.float64 e np.int64
    data['temperatura'] = np.array(data['temperatura'], dtype=np.float32)
    data['velocidade'] = np.array(data['velocidade'], dtype=np.int32)
    ts = TimeSerie(data, format='wide', features_qty=len(data))
    df = ts.df
    cols_str = [name for name in sorted(df.columns)]
    print('cols_str :', ', '.join(cols_str))
    print(f'Dataframe com {len(df.columns)} colunas: {df.columns}')
    # Imprime a série temporal
    print(ts)
    # Grava a série temporal em parquet
    print(f'Grava a série temporal (formato {ts.format}) em um arquivo parquet {path}')
    context = TSWriter(WriteParquetFile())
    print("Client: Strategy was seted to write Parquet file.")
    context.write(Path(path_str), ts)
    print(f'\nArquivo {str(path)} gerado à partir da TimeSerie fornecida')



from pathlib import Path
from datetime import datetime
import numpy as np
import pandas as pd
from t8s.ts import TimeSerie
from t8s.ts_builder import TSBuilder
from t8s.ts_builder import ReadParquetFile

def test_build_from_file():
    path_str: str = 'ts_01.parquet'
    path = Path(path_str)
    print('path: ', path)
    context = TSBuilder(ReadParquetFile())
    print("Client: Strategy is set to read Parquet file.")
    ts: TimeSerie = context.build_from_file(Path(path_str))
    assert int(ts.features) == 3
    assert ts.format == 'wide'
    assert ts.df.__len__() == 4
     # pd._libs.tslibs.timestamps.Timestamp é privado e devo usar pd.Timestamp
    assert type(ts.df['timestamp'][0]) == pd.Timestamp
    assert type(ts.df['temperatura'][0]) == np.float32
    assert type(ts.df['velocidade'][0]) == np.int32


import logging
from logging import Logger
from typing import Any

class LogConfigMeta(type):
    """
    The Singleton class can be implemented in different ways in Python. Some possible methods include: 
    base class, decorator, metaclass. We will use the metaclass because it is best suited for this purpose.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]

class LogConfig(metaclass=LogConfigMeta):
    def __init__(self) -> None:
        self.logger = None
        self.initialize_logger(level=logging.DEBUG)

    def initialize_logger(self, level=logging.DEBUG):
        my_global_logger = logging.getLogger(__name__)
        my_global_logger.setLevel(level)
        #logger_handler = RotatingFileHandler('timeseries.log', maxBytes=1_000_000, backupCount=10)
        logger_handler = logging.FileHandler('timeseries.log', mode='w')
        logger_handler.setFormatter(logging.Formatter("%(asctime)s - %(levelname)s - %(filename)s - %(funcName)s | %(message)s")) #"%(asctime)s - %(name)s - %(levelname)s - %(filename)s:%(funcName)s(%(lineno)d) - %(message)s"
        my_global_logger.handlers.clear()
        my_global_logger.addHandler(logger_handler)
        self.logger = my_global_logger
        self.logger.info('self.logger = ' + str(self.logger))

    def getLogger(self) -> Logger | Any :
        return self.logger



from pathlib import Path
from datetime import datetime
import yaml
from abc import ABC, abstractmethod
from typing import Any, Optional

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

from t8s.log_config import LogConfig

logger = LogConfig().getLogger()

# ts = pd.read_parquet('lixo.parquet', engine='pyarrow')

# Classe abstrata que define a interface para os processamentos de séries temporais
class ITimeSeriesProcessor(ABC):
    def __init__(self):
        self.format: str
        self.features: str
        self.df: pd.DataFrame

    """Converte a série temporal para o formato Long"""
    @abstractmethod
    def to_long(self) -> bool:
        return False

    @abstractmethod
    def to_wide(self) -> bool:
        # Converte a série temporal para o formato Wide
        return False

    @abstractmethod
    def is_univariate(self) -> bool:
        # Verifica se a série temporal é univariada
        return False

    @abstractmethod
    def is_multivariate(self) -> bool:
        # Verifica se a série temporal é multivariada
        return False

    @abstractmethod
    def split(self) -> list[Optional['TimeSerie']]:
        # Cria várias séries temporais univariadas à partir de uma série temporal multivariada
        pass

"""Rastreia a proveniência dos ativos de informação, Série Temporal neste caso"""
class IProvenancable(ABC):
    def __init__(self):
        self.format: str
        self.features: str
        self.df: pd.DataFrame

    """Este método adiciona informações de proveniência ao objeto TimeSerie para 
    uma transformação específica. Ele recebe como parâmetros o nome da transformação 
    e um dicionário com os parâmetros usados na transformação."""
    @abstractmethod
    def add_provenance(self, transformation: str, parameters: dict): 
        pass 
    
    @abstractmethod
    def get_provenances(self) -> list[dict[str, Any]]: 
        """Este método retorna uma lista de dicionários com as informações de 
        proveniência de todas as transformações realizadas no objeto TimeSerie."""
        pass

    @abstractmethod
    def get_provenance_by_transformation(self, transformation: str) -> dict[str, Any]: 
        """Este método retorna as informações de proveniência de uma dada transformação 
        realizadas no objeto TimeSerie."""
        pass

    @abstractmethod
    def get_last_transformation(self, transformation: str) -> dict[str, Any]: 
        """Este método retorna informações da ultima transformação realizada no objeto TimeSerie."""
        pass

    @abstractmethod
    def apply_transformation(self, transformation: str, **kwargs) -> Any : 
        """Este método aplica uma transformação ao objeto TimeSerie e retorna um novo 
        objeto TimeSerie com as informações de proveniência atualizadas. Ele recebe como 
        parâmetros o nome da transformação e os parâmetros específicos da transformação."""
        pass

class ITimeSerie(ITimeSeriesProcessor, IProvenancable):
    """
    Interface que estende as outras interfaces. ë apenas uma Marker Interface
    """
    pass

class TimeSerie(ITimeSerie):
    def __init__(self, *args, format, features_qty, **kwargs):
        assert isinstance(format, str), "format must be a string"
        assert isinstance(features_qty, int), "features_qty must be a int"
        assert format in ['long', 'wide'], "format must be 'long' or 'wide'"
        self.format: str = format
        self.features: str = str(features_qty)
        self.df = pd.DataFrame(*args, **kwargs)

    def __repr__(self):
        # Retorna uma representação em string do objeto TimeSerie
        return f'TimeSerie(format={self.format}, features={self.features}, df=\n{self.df.__repr__()})'
        
    def to_long(self):
        # Converte a série temporal para o formato Long
        # Implementação aqui
        # TODO: Implementar to_long()
        pass

    def to_wide(self):
        # Converte a série temporal para o formato Wide
        # Implementação aqui
        # TODO: Implementar to_wide()
        pass

    def is_univariate(self):
        # Verifica se a série temporal é univariada
        return self.df.columns.size == 2

    def is_multivariate(self):
        # Verifica se a série temporal é multivariada
        return self.df.columns.size > 2

    def write_parquet_file(self, path):
        # Grava os dados em formato Parquet com metadados do objeto TimeSerie
        # to_parquet(path, df, self.format, self.features)
        table = pa.Table.from_pandas(self.df)
        # table = table.replace_schema_metadata({'format': self.format, 'features': self.features})
        table = table.replace_schema_metadata({
            b'format': str(self.format).encode(),
            b'features': str(self.features).encode()})
        pq.write_table(table, path)

    def split(self) -> list[Optional['TimeSerie']]:
        # Cria várias séries temporais univariadas à partir de uma série temporal multivariada
        result = []
        if self.format == 'long':
            # TODO: Implementar para o caso de formato longo.  
            pass
        elif self.format == 'wide':
            # Por contrato a primeira coluna é sempre o timestamp
            timestamp_col = self.df.columns[0]
            for idx, col in enumerate(self.df.columns):
                if idx == 0:
                    continue
                my_df = self.df[[timestamp_col, col]]
                # TODO: criar um novo objeto TimeSerie 
                my_ts = TimeSerie(data=my_df, format='wide', features_qty=2)
                logger.debug('---------------------------------------------------')
                logger.debug(f'univariate {idx-1}', '\n ', my_df)
                result.append(my_ts)
        else:
            raise Exception('Formato de série temporal não suportado')

        msg = 'O método split deve retornar uma lista de objetos TimeSerie'
        for ts in result:
            assert isinstance(ts, TimeSerie), msg

        return result

    ### Métodos de IProvenancable
    def add_provenance(self, transformation: str, parameters: dict):
        # Este método adiciona informações de proveniência ao objeto TimeSerie para 
        # uma transformação específica. Ele recebe como parâmetros o nome da transformação 
        # e um dicionário com os parâmetros usados na transformação.
        # TODO: Implementar add_provenance()
        pass

    def get_provenances(self) -> list[dict[str, Any]]:
        # Este método retorna uma lista de dicionários com as informações de 
        # proveniência de todas as transformações realizadas no objeto TimeSerie.
        # TODO: Implementar get_provenances()
        return []

    def get_provenance_by_transformation(self, transformation: str) -> dict[str, Any]:
        # Este método retorna as informações de proveniência de uma dada transformação 
        # realizadas no objeto TimeSerie.
        # TODO: Implementar get_provenance_by_transformation()
        return {}

    def get_last_transformation(self, transformation: str) -> dict[str, Any]:
        # TODO: Implementar get_provenance_by_transformation()
        return {}

    def apply_transformation(self, transformation: str, **kwargs) -> Optional['TimeSerie']: 
        """Este método aplica uma transformação ao objeto TimeSerie e retorna um novo 
        objeto TimeSerie com as informações de proveniência atualizadas. Ele recebe como 
        parâmetros o nome da transformação e os parâmetros específicos da transformação."""
        # TODO: Implementar apply_transformation()
        return TimeSerie.empty()
    
    @staticmethod
    def empty() -> Any:
        return TimeSerie(format='long', features_qty=0, data={})
